use actix_web::{
    client::ClientRequest, error::BlockingError, http::header::InvalidHeaderValue, web,
};
use std::{fmt::Display, future::Future, pin::Pin};

use crate::{create::Signed, Config, PrepareSignError, Sign};

impl Sign for ClientRequest {
    fn authorization_signature<F, E, K>(
        mut self,
        config: Config,
        key_id: K,
        f: F,
    ) -> Pin<Box<dyn Future<Output = Result<Self, E>>>>
    where
        F: FnOnce(&str) -> Result<String, E> + Send + 'static,
        E: From<BlockingError<E>>
            + From<PrepareSignError>
            + From<InvalidHeaderValue>
            + std::fmt::Debug
            + Send
            + 'static,
        K: Display + 'static,
        Self: Sized,
    {
        Box::pin(async move {
            let signed = prepare(&self, &config, key_id, f).await?;
            signed.authorization_header(self.headers_mut())?;
            Ok(self)
        })
    }

    fn signature<F, E, K>(
        mut self,
        config: Config,
        key_id: K,
        f: F,
    ) -> Pin<Box<dyn Future<Output = Result<Self, E>>>>
    where
        F: FnOnce(&str) -> Result<String, E> + Send + 'static,
        E: From<BlockingError<E>>
            + From<PrepareSignError>
            + From<InvalidHeaderValue>
            + std::fmt::Debug
            + Send
            + 'static,
        K: Display + 'static,
        Self: Sized,
    {
        Box::pin(async move {
            let signed = prepare(&self, &config, key_id, f).await?;
            signed.signature_header(self.headers_mut())?;
            Ok(self)
        })
    }
}

async fn prepare<F, E, K>(
    request: &ClientRequest,
    config: &Config,
    key_id: K,
    f: F,
) -> Result<Signed, E>
where
    F: FnOnce(&str) -> Result<String, E> + Send + 'static,
    E: From<BlockingError<E>> + From<PrepareSignError> + std::fmt::Debug + Send + 'static,
    K: Display,
{
    let unsigned = config.begin_sign(
        request.get_method(),
        request.get_uri().path_and_query(),
        request.headers().clone(),
    )?;

    let key_id = key_id.to_string();

    let signed = web::block(move || unsigned.sign(key_id, f)).await?;

    Ok(signed)
}
